package tech.krave.twilio_chat_plugin

import android.app.Activity
import android.content.Context
import com.twilio.chat.*
import io.flutter.embedding.engine.plugins.activity.ActivityAware
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel


class PluginHandler : MethodChannel.MethodCallHandler, ActivityAware {
    private var applicationContext: Context

    private var activity: Activity? = null

    private val CLIENT_IS_NULL_CODE = "CLIENT_IS_NULL"
    private val CLIENT_IS_NULL_MESSAGE = "Client has not been initialized"

    private val ARGUMENT_IS_NOT_NOT_PROVIDED_CODE = "ARGUMENT_IS_NOT_NOT_PROVIDED"
    private val ARGUMENT_IS_NOT_NOT_PROVIDED_MESSAGE = "One of the arguments is missing"

    @Suppress("ConvertSecondaryConstructorToPrimary")
    constructor(applicationContext: Context) {
        this.applicationContext = applicationContext
    }

    override fun onMethodCall(call: MethodCall, result: MethodChannel.Result) {
        TwilioChatPlugin.debug("PluginHandler.onMethodCall => received ${call.method}")
        when (call.method) {
            "debug" -> debug(call, result)
            "createChatClient" -> createChatClient(call, result)
            "updateUserFriendlyName" -> updateUserFriendlyName(call, result)
            "registerFCMToken" -> registerFCMToken(call, result)
            "unregisterFCMToken" -> unregisterFCMToken(call, result)
            "replaceFCMToken" -> replaceFCMToken(call, result)
            "enterChannel" -> enterChannel(call, result)
            "exitChannel" -> exitChannel(call, result)
            "listPublicChannels" -> listPublicChannels(result)
            "listUserChannels" -> listUserChannels(result)
            "createChannel" -> createChannel(call, result)
            "joinChannel" -> joinChannel(call, result)
            "leaveChannel" -> leaveChannel(call, result)
            "sendMessagesByChannelSID" -> sendMessagesByChannelSID(call, result)
            "getMostRecentMessagesByChannelSID" -> getMostRecentMessagesByChannelSID(call, result)
            "getMessageBefore" -> getMessagesBefore(call, result)
            "inviteUserToChannel" -> inviteUserToChannel(call, result)
            "updateChannelFriendlyName" -> updateChannelFriendlyName(call, result)
            "createPrivateChannelByIdentites" -> createPrivateChannelByIdentites(call, result)
        }
    }

    override fun onDetachedFromActivity() {
        this.activity = null
    }

    override fun onReattachedToActivityForConfigChanges(binding: ActivityPluginBinding) {
        this.activity = binding.activity
    }

    override fun onAttachedToActivity(binding: ActivityPluginBinding) {
        this.activity = binding.activity
    }

    override fun onDetachedFromActivityForConfigChanges() {
        this.activity = null
    }

    private fun debug(call: MethodCall, result: MethodChannel.Result) {
        val enableNative = call.argument<Boolean>("native")
        if (enableNative != null) {
            TwilioChatPlugin.nativeDebug = enableNative
            result.success(enableNative)
        } else {
            result.error("MISSING_PARAMS", "Missing 'native' parameter", null)
        }
    }

    private fun createChatClient(call: MethodCall, result: MethodChannel.Result) {
        val accessToken = call.argument<String>("accessToken").orEmpty()
        if (TwilioChatPlugin.chatClient != null) {
            TwilioChatPlugin.chatClient!!.updateToken(accessToken, object : StatusListener() {
                override fun onSuccess() {
                    return result.success(SerializationHelper.chatClientToMap(TwilioChatPlugin.chatClient))
                }
                override fun onError(errorInfo: ErrorInfo?) {
                    TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                    result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                    super.onError(errorInfo)
                }
            })
        } else {
            val props = ChatClient.Properties.Builder()
                    .setRegion("us1")
                    .createProperties()

            ChatClient.create(
                    applicationContext,
                    accessToken,
                    props,
                    object : CallbackListener<ChatClient>() {
                        override fun onSuccess(chatClient: ChatClient?) {
                            chatClient?.removeListener();
                            if(TwilioChatPlugin.chatClientListener != null){
                                chatClient?.setListener(TwilioChatPlugin.chatClientListener)
                            }
                            TwilioChatPlugin.chatClient = chatClient
                            return result.success(SerializationHelper.chatClientToMap(chatClient))
                        }

                        override fun onError(errorInfo: ErrorInfo?) {
                            super.onError(errorInfo)
                            TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                            return result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                        }
                    })
        }


    }

    private fun updateUserFriendlyName(call: MethodCall, result: MethodChannel.Result) {
        if (TwilioChatPlugin.chatClient == null) {
            return result.error(CLIENT_IS_NULL_CODE, CLIENT_IS_NULL_MESSAGE, null)
        }

        val friendlyName = call.argument<String>("friendlyName")
                ?: return result.error(ARGUMENT_IS_NOT_NOT_PROVIDED_CODE, ARGUMENT_IS_NOT_NOT_PROVIDED_MESSAGE, null)

        TwilioChatPlugin.chatClient?.users?.myUser?.setFriendlyName(friendlyName, object: StatusListener(){
            override fun onSuccess() {
                result.success(true)
            }

            override fun onError(errorInfo: ErrorInfo?) {
                TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                super.onError(errorInfo)
            }
        })
    }

    private fun enterChannel(call: MethodCall, result: MethodChannel.Result) {
        if (TwilioChatPlugin.chatClient == null) {
            return result.error(CLIENT_IS_NULL_CODE, CLIENT_IS_NULL_MESSAGE, null)
        }

        val channelSID = call.argument<String>("channelSID")
                ?: return result.error(ARGUMENT_IS_NOT_NOT_PROVIDED_CODE, ARGUMENT_IS_NOT_NOT_PROVIDED_MESSAGE, null)

        TwilioChatPlugin.chatClient?.channels?.getChannel(channelSID, object : CallbackListener<Channel>() {
            override fun onSuccess(channel: Channel) {
                channel.addListener(TwilioChatPlugin.channelListener)
                TwilioChatPlugin.channel = channel
                result.success(SerializationHelper.channelToMap(channel))
            }
            override fun onError(errorInfo: ErrorInfo?) {
                TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                super.onError(errorInfo)
            }
        })
    }

    private fun exitChannel(call: MethodCall, result: MethodChannel.Result) {
        if (TwilioChatPlugin.chatClient == null) {
            return result.error(CLIENT_IS_NULL_CODE, CLIENT_IS_NULL_MESSAGE, null)
        }
        if (TwilioChatPlugin.channel == null) {
            return result.error(CLIENT_IS_NULL_CODE, CLIENT_IS_NULL_MESSAGE, null)
        }
        TwilioChatPlugin.channel?.removeAllListeners()
        TwilioChatPlugin.channel = null
        result.success(true)
    }

    private fun registerFCMToken(call: MethodCall, result: MethodChannel.Result) {
        if (TwilioChatPlugin.chatClient == null) {
            return result.error(CLIENT_IS_NULL_CODE, CLIENT_IS_NULL_MESSAGE, null)
        }

        val fcmToken = call.argument<String>("fcmToken")
                ?: return result.error(ARGUMENT_IS_NOT_NOT_PROVIDED_CODE, ARGUMENT_IS_NOT_NOT_PROVIDED_MESSAGE, null)


        TwilioChatPlugin.chatClient?.registerFCMToken(ChatClient.FCMToken(fcmToken), object : StatusListener() {
            override fun onSuccess() {
                result.success(SerializationHelper.chatClientToMap(TwilioChatPlugin.chatClient))
            }

            override fun onError(errorInfo: ErrorInfo?) {
                TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                super.onError(errorInfo)
            }

        })

    }

    private fun unregisterFCMToken(call: MethodCall, result: MethodChannel.Result) {
        if (TwilioChatPlugin.chatClient == null) {
            return result.error(CLIENT_IS_NULL_CODE, CLIENT_IS_NULL_MESSAGE, null)
        }

        val fcmToken = call.argument<String>("fcmToken")
                ?: return result.error(ARGUMENT_IS_NOT_NOT_PROVIDED_CODE, ARGUMENT_IS_NOT_NOT_PROVIDED_MESSAGE, null)

        TwilioChatPlugin.chatClient?.unregisterFCMToken(ChatClient.FCMToken(fcmToken), object : StatusListener() {
            override fun onSuccess() {
                result.success(SerializationHelper.chatClientToMap(TwilioChatPlugin.chatClient))
            }

            override fun onError(errorInfo: ErrorInfo?) {
                TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                super.onError(errorInfo)
            }

        })
    }

    private fun replaceFCMToken(call: MethodCall, result: MethodChannel.Result) {
        if (TwilioChatPlugin.chatClient == null) {
            return result.error(CLIENT_IS_NULL_CODE, CLIENT_IS_NULL_MESSAGE, null)
        }

        val oldFcmToken = call.argument<String>("oldFcmToken")
        val newFcmToken = call.argument<String>("newFcmToken")

        if(oldFcmToken == null || newFcmToken == null){
            return result.error(ARGUMENT_IS_NOT_NOT_PROVIDED_CODE, ARGUMENT_IS_NOT_NOT_PROVIDED_MESSAGE, null)
        }

        TwilioChatPlugin.chatClient?.unregisterFCMToken(ChatClient.FCMToken(oldFcmToken), object : StatusListener() {
            override fun onSuccess() {
                TwilioChatPlugin.chatClient?.registerFCMToken(ChatClient.FCMToken(newFcmToken), object : StatusListener() {
                    override fun onSuccess() {
                        result.success("")
                    }

                    override fun onError(errorInfo: ErrorInfo?) {
                        super.onError(errorInfo)
                        return result.error(errorInfo?.code.toString(), errorInfo?.message, null)
                    }
                })
            }

            override fun onError(errorInfo: ErrorInfo?) {
                TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                super.onError(errorInfo)
            }
        })

    }

    private fun listPublicChannels(result: MethodChannel.Result) {
        if (TwilioChatPlugin.chatClient == null) {
            return result.error(CLIENT_IS_NULL_CODE, CLIENT_IS_NULL_MESSAGE, null)
        }

        TwilioChatPlugin.chatClient?.channels?.getPublicChannelsList(object : CallbackListener<Paginator<ChannelDescriptor>>() {
            override fun onSuccess(channelPaginator: Paginator<ChannelDescriptor>) {
                TwilioChatPlugin.debug("ChatClient.channels.getPublicChannelsList => Success")
                val channelList: ArrayList<Map<String, Any?>> = arrayListOf()
                TwilioChatPlugin.publicChannelDescriptors = arrayListOf()
                for (channel in channelPaginator.items) {
                    TwilioChatPlugin.debug("Public channel => ${channel.sid}")
                    val channelMap = SerializationHelper.channelDescriptorToMap(channel)
                    channelList.add(channelMap)
                    TwilioChatPlugin.publicChannelDescriptors?.add(channel)
                }

                result.success(channelList)
            }
            override fun onError(errorInfo: ErrorInfo?) {
                TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                super.onError(errorInfo)
            }
        })
    }

    private fun listUserChannels(result: MethodChannel.Result) {
        if (TwilioChatPlugin.chatClient == null) {
            return result.error(CLIENT_IS_NULL_CODE, CLIENT_IS_NULL_MESSAGE, null)
        }

        TwilioChatPlugin.chatClient?.channels?.getUserChannelsList(object : CallbackListener<Paginator<ChannelDescriptor>>() {
            override fun onSuccess(channelPaginator: Paginator<ChannelDescriptor>) {
                TwilioChatPlugin.debug("ChatClient.channels.getPublicChannelsList => Success")
                val channelList: ArrayList<Map<String, Any?>> = arrayListOf()
                TwilioChatPlugin.userChannelDescriptors = arrayListOf()
                for (channel in channelPaginator.items) {
                    TwilioChatPlugin.debug("Public channel => ${channel.sid}")
                    val channelMap = SerializationHelper.channelDescriptorToMap(channel)
                    channelList.add(channelMap)
                    TwilioChatPlugin.userChannelDescriptors?.add(channel)
                }
                result.success(channelList)
            }
            override fun onError(errorInfo: ErrorInfo?) {
                TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                super.onError(errorInfo)
            }
        })
    }

    private fun createChannel(call: MethodCall, result: MethodChannel.Result) {
        val friendlyName = call.argument<String>("friendlyName")
        val channelTypeInt = call.argument<Int>("channelType")

        if(!(friendlyName != null && channelTypeInt != null)){
            return result.error(ARGUMENT_IS_NOT_NOT_PROVIDED_CODE, ARGUMENT_IS_NOT_NOT_PROVIDED_MESSAGE, null)
        }

        if (TwilioChatPlugin.chatClient == null) {
            return result.error(CLIENT_IS_NULL_CODE, CLIENT_IS_NULL_MESSAGE, null)
        }

        TwilioChatPlugin.chatClient?.channels?.channelBuilder()
                ?.withFriendlyName(friendlyName)
                ?.withType(Channel.ChannelType.fromInt(channelTypeInt))
                ?.build(object : CallbackListener<Channel>() {
                    override fun onSuccess(channel: Channel) {
                        channel.join(object : StatusListener() {
                            override fun onSuccess() {
                                result.success(SerializationHelper.channelToMap(channel))
                            }

                            override fun onError(errorInfo: ErrorInfo?) {
                                result.error(errorInfo?.code.toString(), errorInfo?.message, null)
                            }
                        })
                    }

                    override fun onError(errorInfo: ErrorInfo?) {
                        TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                        result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                        super.onError(errorInfo)
                    }
                })
    }

    private fun createPrivateChannelByIdentites(call: MethodCall, result: MethodChannel.Result) {
        val userIdentities = call.argument<List<String>>("userIdentities")
                ?: return result.error(ARGUMENT_IS_NOT_NOT_PROVIDED_CODE, ARGUMENT_IS_NOT_NOT_PROVIDED_MESSAGE, null)

        if (TwilioChatPlugin.chatClient == null) {
            return result.error(CLIENT_IS_NULL_CODE, CLIENT_IS_NULL_MESSAGE, null)
        }
        TwilioChatPlugin.chatClient?.channels?.channelBuilder()
                ?.withFriendlyName("membersNames")
                ?.withType(Channel.ChannelType.PRIVATE)
                ?.build(object : CallbackListener<Channel>() {
                    override fun onSuccess(channel: Channel) {
                        channel.join(object : StatusListener() {
                            override fun onSuccess() {
                                for (userIdentity in userIdentities){
                                    channel.members.addByIdentity(userIdentity, object: StatusListener(){
                                        override fun onSuccess() {}
                                        override fun onError(errorInfo: ErrorInfo?) {
                                            super.onError(errorInfo)
                                            result.error(errorInfo?.code.toString(), errorInfo?.message, null)
                                        }

                                    })
                                }
                                result.success(SerializationHelper.channelToMap(channel))
                            }

                            override fun onError(errorInfo: ErrorInfo?) {
                                result.error(errorInfo?.code.toString(), errorInfo?.message, null)
                            }
                        })
                    }

                    override fun onError(errorInfo: ErrorInfo?) {
                        TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                        result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                        super.onError(errorInfo)
                    }
                })
    }

    private fun joinChannel(call: MethodCall, result: MethodChannel.Result) {
        val channelSID = call.argument<String>("channelSID")
                ?: return result.error(ARGUMENT_IS_NOT_NOT_PROVIDED_CODE, ARGUMENT_IS_NOT_NOT_PROVIDED_MESSAGE, null)

        if (TwilioChatPlugin.chatClient == null) {
            return result.error(CLIENT_IS_NULL_CODE, CLIENT_IS_NULL_MESSAGE, null)
        }

        TwilioChatPlugin.chatClient?.channels?.getChannel(channelSID, object : CallbackListener<Channel>() {
            override fun onSuccess(channel: Channel) {
                channel.join(object : StatusListener() {
                    override fun onSuccess() {
                        result.success(SerializationHelper.channelToMap(channel))
                    }

                    override fun onError(errorInfo: ErrorInfo?) {
                        TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                        result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                        super.onError(errorInfo)
                    }
                })
            }
        })
    }

    private fun leaveChannel(call: MethodCall, result: MethodChannel.Result) {
        val channelSID = call.argument<String>("channelSID")
                ?: return result.error(ARGUMENT_IS_NOT_NOT_PROVIDED_CODE, ARGUMENT_IS_NOT_NOT_PROVIDED_MESSAGE, null)

        if (TwilioChatPlugin.chatClient == null) {
            return result.error(CLIENT_IS_NULL_CODE, CLIENT_IS_NULL_MESSAGE, null)
        }

        TwilioChatPlugin.chatClient?.channels?.getChannel(channelSID, object : CallbackListener<Channel>() {
            override fun onSuccess(channel: Channel) {
                channel.leave(object : StatusListener() {
                    override fun onSuccess() {
                        result.success(SerializationHelper.channelToMap(channel))
                    }

                    override fun onError(errorInfo: ErrorInfo?) {
                        TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                        result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                        super.onError(errorInfo)
                    }
                })
            }
        })
    }

    private fun sendMessagesByChannelSID(call: MethodCall, result: MethodChannel.Result) {
        val channelSID = call.argument<String>("channelSID")
        val messageBody = call.argument<String>("messageBody")

        if(channelSID == null || messageBody == null){
            return result.error(ARGUMENT_IS_NOT_NOT_PROVIDED_CODE, ARGUMENT_IS_NOT_NOT_PROVIDED_MESSAGE, null)
        }

        if (TwilioChatPlugin.chatClient == null) {
            return result.error(CLIENT_IS_NULL_CODE, CLIENT_IS_NULL_MESSAGE, null)
        }
        if(TwilioChatPlugin.channel != null){
            if(TwilioChatPlugin.channel!!.messages != null){
                TwilioChatPlugin.channel!!.messages.sendMessage(Message.options().withBody(messageBody), object : CallbackListener<Message>() {
                    override fun onSuccess(message: Message?) {
                        result.success(SerializationHelper.messageToMap(message))
                    }
                    override fun onError(errorInfo: ErrorInfo?) {
                        TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                        result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                        super.onError(errorInfo)
                    }
                })
            } else {
                result.error(CLIENT_IS_NULL_CODE, CLIENT_IS_NULL_MESSAGE, "")
            }

        } else {
            TwilioChatPlugin.chatClient?.channels?.getChannel(channelSID, object : CallbackListener<Channel>() {
                override fun onSuccess(channel: Channel) {
                    channel.messages.sendMessage(Message.options().withBody(messageBody), object : CallbackListener<Message>() {
                        override fun onSuccess(message: Message?) {
                            result.success(SerializationHelper.messageToMap(message))
                        }
                    })
                }
            })
        }

    }

    private fun getMostRecentMessagesByChannelSID(call: MethodCall, result: MethodChannel.Result) {
        val channelSID = call.argument<String>("channelSID")
        val messageCount = call.argument<Int>("messageCount")

        if(!(channelSID != null && messageCount != null)){
            return result.error(ARGUMENT_IS_NOT_NOT_PROVIDED_CODE, ARGUMENT_IS_NOT_NOT_PROVIDED_MESSAGE, null)
        }

        if (TwilioChatPlugin.chatClient == null) {
            return result.error(CLIENT_IS_NULL_CODE, CLIENT_IS_NULL_MESSAGE, null)
        }

        if (TwilioChatPlugin.channel != null && channelSID == TwilioChatPlugin.channel!!.sid){
            if(TwilioChatPlugin.channel!!.messages == null){
                val msgList = listOf<Message>()
                return result.success(msgList)
            }
            TwilioChatPlugin.channel!!.messages.getLastMessages(messageCount, object : CallbackListener<List<Message>>() {
                override fun onSuccess(messages: List<Message>?) {
                    val messageList = messages?.map { m -> SerializationHelper.messageToMap(m) }
                    result.success(messageList)
                }
                override fun onError(errorInfo: ErrorInfo?) {
                    TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                    result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                    super.onError(errorInfo)
                }
            })
        } else {
            TwilioChatPlugin.chatClient?.channels?.getChannel(channelSID, object : CallbackListener<Channel>() {
                override fun onSuccess(channel: Channel) {
                    channel.messages.getLastMessages(messageCount, object : CallbackListener<List<Message>>() {
                        override fun onSuccess(messages: List<Message>?) {
                            val messageList = messages?.map { m -> SerializationHelper.messageToMap(m) }
                            result.success(messageList)
                        }

                    })
                }
                override fun onError(errorInfo: ErrorInfo?) {
                    TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                    result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                    super.onError(errorInfo)
                }
            })

        }

    }

    private fun getMessagesBefore(call: MethodCall, result: MethodChannel.Result) {
        val channelSID = call.argument<String>("channelSID")
        val messageCount = call.argument<Int>("messageCount")
        val before = call.argument<Long>("before")

        if(!(channelSID != null && messageCount != null && before != null)){
            return result.error(ARGUMENT_IS_NOT_NOT_PROVIDED_CODE, ARGUMENT_IS_NOT_NOT_PROVIDED_MESSAGE, null)
        }

        if (TwilioChatPlugin.chatClient == null) {
            return result.error(CLIENT_IS_NULL_CODE, CLIENT_IS_NULL_MESSAGE, null)
        }

        if (TwilioChatPlugin.channel != null && channelSID == TwilioChatPlugin.channel!!.sid ){
            if(TwilioChatPlugin.channel!!.messages == null){
                val msgList = listOf<Message>()
                return result.success(msgList)
            }
            TwilioChatPlugin.channel!!.messages.getMessagesBefore(before, messageCount, object : CallbackListener<List<Message>>() {
                override fun onSuccess(messages: List<Message>?) {
                    val messageList = messages?.map { m -> SerializationHelper.messageToMap(m) }
                    result.success(messageList)
                }
                override fun onError(errorInfo: ErrorInfo?) {
                    TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                    result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                    super.onError(errorInfo)
                }
            })
        } else {
            TwilioChatPlugin.chatClient?.channels?.getChannel(channelSID, object : CallbackListener<Channel>() {
                override fun onSuccess(channel: Channel) {
                    channel.messages.getMessagesBefore(before, messageCount, object : CallbackListener<List<Message>>() {
                        override fun onSuccess(messages: List<Message>?) {
                            val messageList = messages?.map { m -> SerializationHelper.messageToMap(m) }
                            result.success(messageList)
                        }
                    })
                }
                override fun onError(errorInfo: ErrorInfo?) {
                    TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                    result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                    super.onError(errorInfo)
                }
            })

        }

    }

    private fun inviteUserToChannel(call: MethodCall, result: MethodChannel.Result){
        val channelSID = call.argument<String>("channelSID")
        val identity = call.argument<String>("identity")

        if(channelSID == null || identity == null ){
            return result.error(ARGUMENT_IS_NOT_NOT_PROVIDED_CODE, ARGUMENT_IS_NOT_NOT_PROVIDED_MESSAGE, null)
        }

        if (TwilioChatPlugin.channel != null && channelSID == TwilioChatPlugin.channel!!.sid){

            TwilioChatPlugin.channel?.members?.inviteByIdentity(identity, object: StatusListener(){
                override fun onSuccess() {
                    return result.success(true)
                }
                override fun onError(errorInfo: ErrorInfo?) {
                    TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                    result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                    super.onError(errorInfo)
                }
            })
        } else {
            TwilioChatPlugin.chatClient?.channels?.getChannel(channelSID, object : CallbackListener<Channel>() {
                override fun onSuccess(channel: Channel) {
                    channel.members?.inviteByIdentity(identity, object: StatusListener(){
                        override fun onSuccess() {
                            return result.success(true)
                        }
                        override fun onError(errorInfo: ErrorInfo?) {
                            TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                            result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                            super.onError(errorInfo)
                        }
                    })
                }
                override fun onError(errorInfo: ErrorInfo?) {
                    TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                    result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                    super.onError(errorInfo)
                }
            })
        }
    }

    private fun updateChannelFriendlyName(call: MethodCall, result: MethodChannel.Result){
        val channelSID = call.argument<String>("channelSID")
        val friendlyName = call.argument<String>("friendlyName")

        if(channelSID == null || friendlyName == null ){
            return result.error(ARGUMENT_IS_NOT_NOT_PROVIDED_CODE, ARGUMENT_IS_NOT_NOT_PROVIDED_MESSAGE, null)
        }

        if (TwilioChatPlugin.channel != null && channelSID == TwilioChatPlugin.channel!!.sid){

            TwilioChatPlugin.channel?.setFriendlyName(friendlyName, object: StatusListener(){
                override fun onSuccess() {
                    return result.success(true)
                }
                override fun onError(errorInfo: ErrorInfo?) {
                    TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                    result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                    super.onError(errorInfo)
                }
            })
        } else {
            TwilioChatPlugin.chatClient?.channels?.getChannel(channelSID, object : CallbackListener<Channel>() {
                override fun onSuccess(channel: Channel) {
                    channel.setFriendlyName(friendlyName, object: StatusListener(){
                        override fun onSuccess() {
                            return result.success(true)
                        }
                        override fun onError(errorInfo: ErrorInfo?) {
                            TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                            result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                            super.onError(errorInfo)
                        }
                    })
                }
                override fun onError(errorInfo: ErrorInfo?) {
                    TwilioChatPlugin.debug("Error => ${errorInfo?.code} ${errorInfo?.message}")
                    result.error(errorInfo?.code.toString(), errorInfo?.message, "")
                    super.onError(errorInfo)
                }
            })
        }
    }


}